@@ -1,8 +1,12 @@
+use std::collections::{HashMap, HashSet};
+use std::ops::{Add, Div, Sub};
 use std::{error::Error, io::Write, ops::Mul};
 
-use binread::BinReaderExt;
 use binread::{io::Cursor, BinRead};
+use binread::{BinReaderExt, BinResult};
 use half::f16;
+use itertools::Itertools;
+use ssbh_lib::formats::mesh::RiggingType;
 use ssbh_lib::{
     formats::mesh::{
         AttributeDataType, AttributeDataTypeV8, AttributeUsageV10, AttributeUsageV8,
@@ -11,7 +15,7 @@ use ssbh_lib::{
     },
     SsbhByteBuffer,
 };
-use ssbh_lib::{Half, SsbhArray};
+use ssbh_lib::{Half, Matrix3x3, SsbhArray, Vector3};
 
 use crate::{read_data, read_vector_data};
 
@@ -35,6 +39,61 @@ pub enum AttributeUsage {
     ColorSet,
 }
 
+/// An error while reading mesh attribute data.
+pub enum AttributeError {
+    /// Attempted to read from a nonexistent buffer.
+    InvalidBufferIndex(u64),
+
+    /// Failed to find the offset or stride in bytes for the given buffer index.
+    NoOffsetOrStride(u64),
+
+    /// An error occurred while reading the data from the buffer.
+    Io(std::io::Error),
+
+    /// An error occurred while reading the data from the buffer.
+    BinRead(binread::error::Error),
+}
+
+impl std::error::Error for AttributeError {}
+
+impl From<std::io::Error> for AttributeError {
+    fn from(e: std::io::Error) -> Self {
+        Self::Io(e)
+    }
+}
+
+impl From<binread::error::Error> for AttributeError {
+    fn from(e: binread::error::Error) -> Self {
+        match e {
+            binread::Error::Io(io) => Self::Io(io),
+            _ => Self::BinRead(e),
+        }
+    }
+}
+
+impl std::fmt::Display for AttributeError {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        std::fmt::Debug::fmt(self, f)
+    }
+}
+
+impl std::fmt::Debug for AttributeError {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            AttributeError::InvalidBufferIndex(index) => {
+                write!(f, "No buffer found for index {}.", index)
+            }
+            AttributeError::NoOffsetOrStride(index) => write!(
+                f,
+                "Found index {}. Buffer indices higher than 1 are not supported.",
+                index
+            ),
+            AttributeError::Io(err) => write!(f, "IO Error: {:?}", err),
+            AttributeError::BinRead(err) => write!(f, "BinRead Error: {:?}", err),
+        }
+    }
+}
+
 impl From<AttributeUsageV10> for AttributeUsage {
     fn from(a: AttributeUsageV10) -> Self {
         match a {
@@ -90,10 +149,7 @@ impl From<AttributeDataTypeV8> for DataType {
     }
 }
 
-fn read_vertex_indices(
-    mesh_index_buffer: &[u8],
-    mesh_object: &MeshObject,
-) -> Result<Vec<u32>, Box<dyn Error>> {
+fn read_vertex_indices(mesh_index_buffer: &[u8], mesh_object: &MeshObject) -> BinResult<Vec<u32>> {
     // Use u32 regardless of the actual data type to simplify conversions.
     let count = mesh_object.vertex_index_count as usize;
     let offset = mesh_object.index_buffer_offset as u64;
@@ -118,26 +174,15 @@ fn read_attribute_data<T>(
     mesh: &Mesh,
     mesh_object: &MeshObject,
     attribute: &MeshAttribute,
-) -> Result<VectorData, Box<dyn Error>> {
+) -> Result<VectorData, AttributeError> {
     // Get the raw data for the attribute for this mesh object.
     let attribute_buffer = mesh
         .vertex_buffers
         .elements
         .get(attribute.index as usize)
-        .ok_or("Invalid buffer index.")?;
+        .ok_or(AttributeError::InvalidBufferIndex(attribute.index))?;
 
-    // TODO: Create functions for this?
-    let offset = match attribute.index {
-        0 => Ok(attribute.offset + mesh_object.vertex_buffer0_offset as u64),
-        1 => Ok(attribute.offset + mesh_object.vertex_buffer1_offset as u64),
-        _ => Err("Buffer indices higher than 1 are not supported."),
-    }? as u64;
-
-    let stride = match attribute.index {
-        0 => Ok(mesh_object.stride0),
-        1 => Ok(mesh_object.stride1),
-        _ => Err("Buffer indices higher than 1 are not supported."),
-    }? as u64;
+    let (offset, stride) = calculate_offset_stride(attribute, mesh_object)?;
 
     let count = mesh_object.vertex_count as usize;
 
@@ -174,17 +219,40 @@ fn read_attribute_data<T>(
             offset,
             stride,
         )?),
-        DataType::Byte4 => VectorData::Vector4(read_vector_data::<_, u8, 4>(
-            &mut reader,
-            count,
-            offset,
-            stride,
-        )?),
+        DataType::Byte4 => {
+            let mut elements = read_vector_data::<_, u8, 4>(&mut reader, count, offset, stride)?;
+            // Normalize the values by converting from the range [0u8, 255u8] to [0.0f32, 1.0f32].
+            for [x, y, z, w] in elements.iter_mut() {
+                *x /= 255f32;
+                *y /= 255f32;
+                *z /= 255f32;
+                *w /= 255f32;
+            }
+            VectorData::Vector4(elements)
+        }
     };
 
     Ok(data)
 }
 
+fn calculate_offset_stride(
+    attribute: &MeshAttribute,
+    mesh_object: &MeshObject,
+) -> Result<(u64, u64), AttributeError> {
+    let (offset, stride) = match attribute.index {
+        0 => Ok((
+            attribute.offset + mesh_object.vertex_buffer0_offset as u64,
+            mesh_object.stride0 as u64,
+        )),
+        1 => Ok((
+            attribute.offset + mesh_object.vertex_buffer1_offset as u64,
+            mesh_object.stride1 as u64,
+        )),
+        _ => Err(AttributeError::NoOffsetOrStride(attribute.index)),
+    }?;
+    Ok((offset, stride))
+}
+
 /// Read data for all attributes of the given `usage` for `mesh_object`.
 pub fn read_attributes_by_usage(
     mesh: &Mesh,
@@ -213,25 +281,8 @@ pub fn read_texture_coordinates(
     for attribute in &get_attributes(&mesh_object, AttributeUsage::TextureCoordinate) {
         let mut data = read_attribute_data::<f32>(mesh, mesh_object, attribute)?;
 
-        // TODO: Clean this up with a function?
         if flip_vertical {
-            match &mut data {
-                VectorData::Vector2(v) => {
-                    for element in v.iter_mut() {
-                        element[1] = 1.0 - element[1];
-                    }
-                }
-                VectorData::Vector3(v) => {
-                    for element in v.iter_mut() {
-                        element[1] = 1.0 - element[1];
-                    }
-                }
-                VectorData::Vector4(v) => {
-                    for element in v.iter_mut() {
-                        element[1] = 1.0 - element[1];
-                    }
-                }
-            }
+            flip_y(&mut data);
         }
 
         attributes.push(AttributeData {
@@ -243,52 +294,33 @@ pub fn read_texture_coordinates(
     Ok(attributes)
 }
 
+fn flip_y(data: &mut VectorData) {
+    match data {
+        VectorData::Vector2(v) => {
+            for [_, y] in v.iter_mut() {
+                *y = 1.0 - *y;
+            }
+        }
+        VectorData::Vector3(v) => {
+            for [_, y, _] in v.iter_mut() {
+                *y = 1.0 - *y;
+            }
+        }
+        VectorData::Vector4(v) => {
+            for [_, y, _, _] in v.iter_mut() {
+                *y = 1.0 - *y;
+            }
+        }
+    }
+}
+
 /// Returns all the colorset attributes for the specified `mesh_object`.
 /// [u8] values are converted to [f32] by normalizing to the range 0.0 to 1.0.
 pub fn read_colorsets(
     mesh: &Mesh,
     mesh_object: &MeshObject,
 ) -> Result<Vec<AttributeData>, Box<dyn Error>> {
-    let colorsets = get_attributes(&mesh_object, AttributeUsage::ColorSet);
-
-    let mut attributes = Vec::new();
-    for attribute in &colorsets {
-        let mut data = read_attribute_data::<f32>(mesh, mesh_object, attribute)?;
-
-        // Normalize integral values by converting the range [0.0, 255.0] to [0.0, 1.0].
-        // TODO: Find a cleaner/safer way to do this.
-        let divisor = 255.0f32;
-        match &mut data {
-            VectorData::Vector2(v) => {
-                for element in v.iter_mut() {
-                    element[0] /= divisor;
-                    element[1] /= divisor;
-                }
-            }
-            VectorData::Vector3(v) => {
-                for element in v.iter_mut() {
-                    element[0] /= divisor;
-                    element[1] /= divisor;
-                    element[2] /= divisor;
-                }
-            }
-            VectorData::Vector4(v) => {
-                for element in v.iter_mut() {
-                    element[0] /= divisor;
-                    element[1] /= divisor;
-                    element[2] /= divisor;
-                    element[3] /= divisor;
-                }
-            }
-        }
-
-        attributes.push(AttributeData {
-            name: attribute.name.clone(),
-            data,
-        });
-    }
-
-    Ok(attributes)
+    read_attributes_by_usage(mesh, mesh_object, AttributeUsage::ColorSet)
 }
 
 fn read_rigging_data(
@@ -330,6 +362,7 @@ pub struct MeshObjectData {
     pub color_sets: Vec<AttributeData>,
     /// Vertex weights grouped by bone name.
     /// Each vertex will likely be influenced by at most 4 bones, but the format doesn't enforce this.
+    /// For single bound objects, [bone_influences](#structfield.bone_influences) should be an empty list.
     pub bone_influences: Vec<BoneInfluence>,
 }
 
@@ -386,55 +419,85 @@ pub fn read_mesh_objects(mesh: &Mesh) -> Result<Vec<MeshObjectData>, Box<dyn Err
     Ok(mesh_objects)
 }
 
-// TODO: Create a new mesh instead.
-pub fn update_mesh(
-    mesh: &mut Mesh,
-    updated_object_data: &[MeshObjectData],
-) -> Result<(), Box<dyn Error>> {
-    let mesh_vertex_data = create_mesh_objects(mesh, updated_object_data)?;
+pub fn create_mesh(
+    major_version: u16,
+    minor_version: u16,
+    object_data: &[MeshObjectData],
+) -> Result<Mesh, Box<dyn Error>> {
+    let mesh_vertex_data = create_mesh_objects(major_version, minor_version, object_data)?;
 
-    mesh.objects.elements = mesh_vertex_data.mesh_objects;
-
-    mesh.vertex_buffers.elements = mesh_vertex_data
-        .vertex_buffers
-        .into_iter()
-        .map(|b| SsbhByteBuffer { elements: b })
+    // TODO: It might be more efficient to reuse the data for mesh object bounding or reuse the generated points.
+    let all_positions: Vec<geometry_tools::glam::Vec3A> = object_data
+        .iter()
+        .map(|o| match o.positions.first() {
+            Some(attribute) => vector_data_to_glam(&attribute.data),
+            None => Vec::new(),
+        })
+        .flatten()
         .collect();
 
-    mesh.index_buffer.elements = mesh_vertex_data.index_buffer;
+    let mesh = Mesh {
+        major_version,
+        minor_version,
+        model_name: "".into(),
+        bounding_info: calculate_bounding_info(&all_positions),
+        unk1: 0,
+        objects: mesh_vertex_data.mesh_objects.into(),
+        // There are always at least 4 buffer entries even if only 2 are used.
+        buffer_sizes: mesh_vertex_data
+            .vertex_buffers
+            .iter()
+            .map(|b| b.len() as u32)
+            .pad_using(4, |_| 0u32)
+            .collect::<Vec<u32>>()
+            .into(),
+        polygon_index_size: mesh_vertex_data.index_buffer.len() as u64,
+        vertex_buffers: mesh_vertex_data
+            .vertex_buffers
+            .into_iter()
+            .map(|b| SsbhByteBuffer::new(b))
+            .collect::<Vec<SsbhByteBuffer>>()
+            .into(),
+        index_buffer: mesh_vertex_data.index_buffer.into(),
+        rigging_buffers: create_rigging_buffers(major_version, minor_version, object_data)?.into(),
+    };
 
-    mesh.rigging_buffers.elements = create_rigging_buffers(mesh, updated_object_data)?;
+    Ok(mesh)
+}
 
-    Ok(())
+fn calculate_max_influences(influences: &[BoneInfluence]) -> usize {
+    // Find the number of influences for the vertex with the most influences.
+    let mut influences_by_vertex = HashMap::new();
+    for influence in influences {
+        for weight in &influence.vertex_weights {
+            // Assume influences are uniquely identified by their bone name.
+            let entry = influences_by_vertex.entry(weight.vertex_index).or_insert(HashSet::new());
+            entry.insert(&influence.bone_name);
+        }
+    }
+
+    influences_by_vertex.values().map(|s| s.len()).max().unwrap_or(0)
 }
 
 fn create_rigging_buffers(
-    source_mesh: &Mesh,
-    updated_object_data: &[MeshObjectData],
+    major_version: u16,
+    minor_version: u16,
+    object_data: &[MeshObjectData],
 ) -> std::io::Result<Vec<MeshRiggingGroup>> {
     let mut rigging_buffers = Vec::new();
 
-    for mesh_object in updated_object_data {
-        // Just use the existing flags for now.
-        // TODO: Properly recreate flags.
-        let flags = match source_mesh.rigging_buffers.elements.iter().find(|r| {
-            r.mesh_object_name.get_string() == Some(&mesh_object.name)
-                && r.mesh_object_sub_index == mesh_object.sub_index
-        }) {
-            Some(r) => r.flags,
-            None => RiggingFlags::new().with_max_influences(4),
+    for mesh_object in object_data {
+        // TODO: unk1 is sometimes set to 0 for singlebound mesh objects, which isn't currently preserved.
+        let flags = RiggingFlags {
+            max_influences: calculate_max_influences(&mesh_object.bone_influences) as u8,
+            unk1: 1,
         };
 
-        //TODO: Find a way to convert &str or &String without an additional clone?
         let mut buffers = Vec::new();
         for i in &mesh_object.bone_influences {
             let buffer = MeshBoneBuffer {
                 bone_name: i.bone_name.clone().into(),
-                data: create_vertex_weights(
-                    source_mesh.major_version,
-                    source_mesh.minor_version,
-                    &i.vertex_weights,
-                )?,
+                data: create_vertex_weights(major_version, minor_version, &i.vertex_weights)?,
             };
             buffers.push(buffer);
         }
@@ -449,7 +512,7 @@ fn create_rigging_buffers(
         rigging_buffers.push(buffer)
     }
 
-    // Rigging buffers need to be sorted in ascending order by name and sub_index. 
+    // Rigging buffers need to be sorted in ascending order by name and sub_index.
     // TODO: Using a default may impact sorting if mesh_object_name is a null offset.
     rigging_buffers.sort_by_key(|k| {
         (
@@ -461,15 +524,14 @@ fn create_rigging_buffers(
     Ok(rigging_buffers)
 }
 
-// TODO: Test both versions.
 fn create_vertex_weights(
     major_version: u16,
     minor_version: u16,
     vertex_weights: &[VertexWeight],
 ) -> std::io::Result<VertexWeights> {
-    // TODO: Create the weights (1.8) or write the buffer (1.10).
     match (major_version, minor_version) {
         (1, 8) => {
+            // Mesh version 1.8 uses an array of structs.
             let weights: Vec<VertexWeightV8> = vertex_weights
                 .iter()
                 .map(|v| VertexWeightV8 {
@@ -480,6 +542,7 @@ fn create_vertex_weights(
             Ok(VertexWeights::VertexWeightsV8(weights.into()))
         }
         (1, 10) => {
+            // Mesh version 1.10 use a byte buffer.
             let mut bytes = Cursor::new(Vec::new());
             for weight in vertex_weights {
                 bytes.write_all(&(weight.vertex_index as u16).to_le_bytes())?;
@@ -487,10 +550,13 @@ fn create_vertex_weights(
             }
             Ok(VertexWeights::VertexWeightsV10(bytes.into_inner().into()))
         }
-        _ => panic!(
-            "Unsupported MESH version {}.{}",
-            major_version, minor_version
-        ),
+        _ => Err(std::io::Error::new(
+            std::io::ErrorKind::Other,
+            format!(
+                "Unsupported MESH version {}.{}",
+                major_version, minor_version
+            ),
+        )),
     }
 }
 
@@ -769,10 +835,12 @@ struct MeshVertexData {
     index_buffer: Vec<u8>,
 }
 
+// TODO: Use a better error type.
 fn create_mesh_objects(
-    source_mesh: &Mesh,
+    major_version: u16,
+    minor_version: u16,
     mesh_object_data: &[MeshObjectData],
-) -> Result<MeshVertexData, Box<dyn Error>> {
+) -> std::io::Result<MeshVertexData> {
     // TODO: Split this into functions and do some cleanup.
     let mut mesh_objects = Vec::new();
 
@@ -784,132 +852,125 @@ fn create_mesh_objects(
     let mut buffer1 = Cursor::new(Vec::new());
 
     for data in mesh_object_data {
-        // This should probably use the existing mesh object data when possible.
-        let source_object = source_mesh
-            .objects
-            .elements
-            .iter()
-            .find(|o| o.name.get_string() == Some(&data.name) && o.sub_index == data.sub_index);
+        let (stride0, stride1, attributes) = create_attributes(data, major_version, minor_version);
 
-        match source_object {
-            Some(source_object) => {
-                let (stride0, stride1, attributes) =
-                    create_attributes(data, source_mesh.major_version, source_mesh.minor_version);
+        // TODO: Make sure all attributes have the same length and return an error if not.
+        // TODO: Allow no position attribute as long as the remaining attributes have equal counts.
+        let position_data = data.positions.get(0).ok_or(std::io::Error::new(
+            std::io::ErrorKind::Other,
+            "Missing position attribute. Failed to determine vertex count.",
+        ))?;
+        let vertex_count = match &position_data.data {
+            VectorData::Vector2(v) => v.len(),
+            VectorData::Vector3(v) => v.len(),
+            VectorData::Vector4(v) => v.len(),
+        } as u32;
 
-                // TODO: Make sure all attributes have the same length.
-                let position_data = data
-                    .positions
-                    .get(0)
-                    .ok_or("Missing position attribute. Failed to determine vertex count.")?;
-                let vertex_count = match &position_data.data {
-                    VectorData::Vector2(v) => v.len(),
-                    VectorData::Vector3(v) => v.len(),
-                    VectorData::Vector4(v) => v.len(),
-                } as u32;
+        // TODO: What does this value do?
+        let unk6 = if major_version == 1 && minor_version == 8 {
+            32
+        } else {
+            0
+        };
 
-                // TODO: What does this value do?
-                let unk6 = if source_mesh.major_version == 1 && source_mesh.minor_version == 8 {
-                    32
-                } else {
-                    0
-                };
+        // Assume generated bounding data isn't critical if there are no points,
+        // so use an empty list if there are no position attributes.
+        let positions = match data.positions.first() {
+            Some(attribute) => vector_data_to_glam(&attribute.data),
+            None => Vec::new(),
+        };
 
-                let mesh_object = MeshObject {
-                    name: data.name.clone().into(),
-                    sub_index: data.sub_index,
-                    parent_bone_name: data.parent_bone_name.clone().into(),
-                    vertex_count: vertex_count as u32,
-                    vertex_index_count: data.vertex_indices.len() as u32,
-                    unk2: 3, // triangles?
-                    vertex_buffer0_offset: buffer0.position() as u32,
-                    vertex_buffer1_offset: buffer1.position() as u32,
-                    final_buffer_offset,
-                    buffer_index: 0, // TODO: This is always 0
-                    stride0,
-                    stride1,
-                    unk6,
-                    unk7: 0,
-                    index_buffer_offset: index_buffer.position() as u32,
-                    unk8: 4,
-                    draw_element_type: source_object.draw_element_type,
-                    rigging_type: source_object.rigging_type,
-                    unk11: 0,
-                    unk12: 0,
-                    bounding_info: source_object.bounding_info, // TODO: Calculate this
-                    attributes,
-                };
+        let mesh_object = MeshObject {
+            name: data.name.clone().into(),
+            sub_index: data.sub_index,
+            parent_bone_name: data.parent_bone_name.clone().into(),
+            vertex_count: vertex_count as u32,
+            vertex_index_count: data.vertex_indices.len() as u32,
+            unk2: 3, // triangles?
+            vertex_buffer0_offset: buffer0.position() as u32,
+            vertex_buffer1_offset: buffer1.position() as u32,
+            final_buffer_offset,
+            buffer_index: 0, // TODO: This is always 0
+            stride0,
+            stride1,
+            unk6,
+            unk7: 0,
+            index_buffer_offset: index_buffer.position() as u32,
+            unk8: 4,
+            draw_element_type: DrawElementType::UnsignedShort,
+            rigging_type: if data.bone_influences.is_empty() {
+                RiggingType::SingleBound
+            } else {
+                RiggingType::Weighted
+            },
+            unk11: 0,
+            unk12: 0,
+            bounding_info: calculate_bounding_info(&positions),
+            attributes,
+        };
 
-                // Assume unsigned short for vertex indices.
-                // TODO: How to handle out of range values?
-                for index in &data.vertex_indices {
-                    index_buffer.write_all(&(*index as u16).to_le_bytes())?;
-                }
+        // Assume unsigned short for vertex indices.
+        // TODO: How to handle values not representable as u16?
+        for index in &data.vertex_indices {
+            index_buffer.write_all(&(*index as u16).to_le_bytes())?;
+        }
 
-                // The first buffer (buffer1) has interleaved data for the required attributes.
-                // Vertex, Data
-                // 0: Position0, Normal0, Tangent0
-                // 1: Position0, Normal0, Tangent0
-                // ...
-                for i in 0..vertex_count as usize {
-                    // TODO: How to write the buffers if the component count isn't yet known?
+        // The first buffer (buffer1) has interleaved data for the required attributes.
+        // Vertex, Data
+        // 0: Position0, Normal0, Tangent0
+        // 1: Position0, Normal0, Tangent0
+        // ...
+        for i in 0..vertex_count as usize {
+            write_all_f32(&mut buffer0, &data.positions, i)?;
 
-                    write_all_f32(&mut buffer0, &data.positions, i)?;
+            // Assume Normal0 and Tangent0 will use half precision.
+            // Binormal uses single precision for some games.
+            // TODO: This won't always match the original (add an option?).
+            write_all_f16(&mut buffer0, &data.normals, i)?;
+            write_all_f32(&mut buffer0, &data.binormals, i)?;
+            write_all_f16(&mut buffer0, &data.tangents, i)?;
+        }
 
-                    // Assume Normal0 and Tangent0 will use half precision.
-                    // Binormal uses single precision for some games.
-                    // TODO: This won't always match the original (add an option?).
-                    write_all_f16(&mut buffer0, &data.normals, i)?;
-                    write_all_f32(&mut buffer0, &data.binormals, i)?;
-                    write_all_f16(&mut buffer0, &data.tangents, i)?;
-                }
+        // The second buffer (buffer1) has interleaved data for the texture coordinates and colorsets.
+        // The attributes differ between meshes, but texture coordinates always precede colorsets.
+        // Vertex, Data
+        // 0: map1, colorSet1
+        // 1: map1, colorSet1
+        // ...
+        // TODO: Make sure all arrays have the same length.
+        for i in 0..vertex_count as usize {
+            // Assume texture coordinates will use half precision.
+            for attribute in &data.texture_coordinates {
+                // TODO: Flipping UVs should be configurable.
 
-                // The second buffer (buffer1) has interleaved data for the texture coordinates and colorsets.
-                // The attributes differ between meshes, but texture coordinates always precede colorsets.
-                // Vertex, Data
-                // 0: map1, colorSet1
-                // 1: map1, colorSet1
-                // ...
-                // TODO: Make sure all arrays have the same length.
-                for i in 0..vertex_count as usize {
-                    // Assume texture coordinates will use half precision.
-                    for attribute in &data.texture_coordinates {
-                        // TODO: Flipping UVs should be configurable.
-
-                        match &attribute.data {
-                            // TODO: Find a better way to handle flipping.
-                            VectorData::Vector2(v) => {
-                                write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
-                            }
-                            VectorData::Vector3(v) => {
-                                write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
-                            }
-                            VectorData::Vector4(v) => {
-                                write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
-                            }
-                        }
+                match &attribute.data {
+                    // TODO: Find a better way to handle flipping.
+                    VectorData::Vector2(v) => {
+                        write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
                     }
-
-                    // Assume u8 for color sets.
-                    for attribute in &data.color_sets {
-                        match &attribute.data {
-                            VectorData::Vector2(v) => write_u8(&mut buffer1, &v[i])?,
-                            VectorData::Vector3(v) => write_u8(&mut buffer1, &v[i])?,
-                            VectorData::Vector4(v) => write_u8(&mut buffer1, &v[i])?,
-                        }
+                    VectorData::Vector3(v) => {
+                        write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
+                    }
+                    VectorData::Vector4(v) => {
+                        write_f16(&mut buffer1, &[v[i][0], 1.0f32 - v[i][1]])?
                     }
                 }
-
-                // TODO: Why is this 32?
-                final_buffer_offset += 32 * mesh_object.vertex_count;
-
-                mesh_objects.push(mesh_object);
             }
-            None => {
-                // TODO: Temporary workaround for not being able to rebuild all the data.
-                // This assumes new mesh objects will not be added and bounding info, rigging info, etc does not change.
-                continue;
+
+            // Assume u8 for color sets.
+            for attribute in &data.color_sets {
+                match &attribute.data {
+                    VectorData::Vector2(v) => write_u8(&mut buffer1, &v[i])?,
+                    VectorData::Vector3(v) => write_u8(&mut buffer1, &v[i])?,
+                    VectorData::Vector4(v) => write_u8(&mut buffer1, &v[i])?,
+                }
             }
         }
+
+        // TODO: Why is this 32?
+        final_buffer_offset += 32 * mesh_object.vertex_count;
+
+        mesh_objects.push(mesh_object);
     }
 
     // There are always four vertex buffers, but only the first two contain data.
@@ -926,7 +987,53 @@ fn create_mesh_objects(
     })
 }
 
-fn write_f32<W: Write>(writer: &mut W, data: &[f32]) -> Result<(), Box<dyn Error>> {
+fn vector_data_to_glam(points: &VectorData) -> Vec<geometry_tools::glam::Vec3A> {
+    match points {
+        VectorData::Vector2(data) => data
+            .iter()
+            .map(|[x, y]| geometry_tools::glam::Vec3A::new(*x, *y, 0f32))
+            .collect(),
+        VectorData::Vector3(data) => data
+            .iter()
+            .map(|[x, y, z]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
+            .collect(),
+        VectorData::Vector4(data) => data
+            .iter()
+            .map(|[x, y, z, _]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
+            .collect(),
+    }
+}
+
+fn calculate_bounding_info(
+    positions: &[geometry_tools::glam::Vec3A],
+) -> ssbh_lib::formats::mesh::BoundingInfo {
+    // Calculate bounding info based on the current points.
+    let (sphere_center, sphere_radius) =
+        geometry_tools::calculate_bounding_sphere_from_points(&positions);
+    let (aabb_min, aabb_max) = geometry_tools::calculate_aabb_from_points(&positions);
+
+    // TODO: Compute a better oriented bounding box.
+    let obb_center = aabb_min.add(aabb_max).div(2f32);
+    let obb_size = aabb_max.sub(aabb_min).div(2f32);
+
+    ssbh_lib::formats::mesh::BoundingInfo {
+        bounding_sphere: ssbh_lib::formats::mesh::BoundingSphere {
+            center: Vector3::new(sphere_center.x, sphere_center.y, sphere_center.z),
+            radius: sphere_radius,
+        },
+        bounding_volume: ssbh_lib::formats::mesh::BoundingVolume {
+            min: Vector3::new(aabb_min.x, aabb_min.y, aabb_min.z),
+            max: Vector3::new(aabb_max.x, aabb_max.y, aabb_max.z),
+        },
+        oriented_bounding_box: ssbh_lib::formats::mesh::OrientedBoundingBox {
+            center: Vector3::new(obb_center.x, obb_center.y, obb_center.z),
+            transform: Matrix3x3::identity(),
+            size: Vector3::new(obb_size.x, obb_size.y, obb_size.z),
+        },
+    }
+}
+
+fn write_f32<W: Write>(writer: &mut W, data: &[f32]) -> std::io::Result<()> {
     for component in data {
         writer.write_all(&component.to_le_bytes())?;
     }
@@ -938,7 +1045,7 @@ fn write_all_f32<W: Write>(
     writer: &mut W,
     attributes: &[AttributeData],
     index: usize,
-) -> Result<(), Box<dyn Error>> {
+) -> std::io::Result<()> {
     for attribute in attributes {
         match &attribute.data {
             VectorData::Vector2(v) => write_f32(writer, &v[index])?,
@@ -949,14 +1056,14 @@ fn write_all_f32<W: Write>(
     Ok(())
 }
 
-fn write_u8<W: Write>(writer: &mut W, data: &[f32]) -> Result<(), Box<dyn Error>> {
+fn write_u8<W: Write>(writer: &mut W, data: &[f32]) -> std::io::Result<()> {
     for component in data {
         writer.write_all(&[get_u8_clamped(*component)])?;
     }
     Ok(())
 }
 
-fn write_f16<W: Write>(writer: &mut W, data: &[f32]) -> Result<(), Box<dyn Error>> {
+fn write_f16<W: Write>(writer: &mut W, data: &[f32]) -> std::io::Result<()> {
     for component in data {
         writer.write_all(&f16::from_f32(*component).to_le_bytes())?;
     }
@@ -967,7 +1074,7 @@ fn write_all_f16<W: Write>(
     writer: &mut W,
     attributes: &[AttributeData],
     index: usize,
-) -> Result<(), Box<dyn Error>> {
+) -> std::io::Result<()> {
     for attribute in attributes {
         match &attribute.data {
             VectorData::Vector2(v) => write_f16(writer, &v[index])?,
@@ -1043,8 +1150,16 @@ impl From<&MeshAttributeV10> for MeshAttribute {
 
 impl From<&MeshAttributeV8> for MeshAttribute {
     fn from(a: &MeshAttributeV8) -> Self {
-        // TODO: Come up with a default name based on usage and subindex.
-        let name = "".to_string();
+        // Version 1.8 doesn't have names.
+        // Generate a name based on conventions for Smash Ultimate and New Pokemon Snap.
+        let name = match a.usage {
+            AttributeUsageV8::Position => format!("Position{}", a.sub_index),
+            AttributeUsageV8::Normal => format!("Normal{}", a.sub_index),
+            AttributeUsageV8::Tangent => format!("Tangent{}", a.sub_index),
+            AttributeUsageV8::TextureCoordinate => format!("TextureCoordinate{}", a.sub_index),
+            AttributeUsageV8::ColorSet => format!("colorSet{}", a.sub_index),
+        };
+
         MeshAttribute {
             name,
             index: a.buffer_index as u64,
@@ -1091,6 +1206,42 @@ mod tests {
         hex::decode(no_whitespace).unwrap()
     }
 
+    #[test]
+    fn attribute_from_attribute_v10() {
+        let attribute_v10 = MeshAttributeV10 {
+            usage: AttributeUsageV10::Normal,
+            data_type: AttributeDataType::HalfFloat2,
+            buffer_index: 2,
+            buffer_offset: 10,
+            sub_index: 3,
+            name: "custom_name".into(),
+            attribute_names: vec!["name1".into()].into(),
+        };
+
+        let attribute: MeshAttribute = (&attribute_v10).into();
+        assert_eq!("name1", attribute.name);
+        assert_eq!(DataType::HalfFloat2, attribute.data_type);
+        assert_eq!(2, attribute.index);
+        assert_eq!(10, attribute.offset);
+    }
+
+    #[test]
+    fn attribute_from_attribute_v8() {
+        let attribute_v8 = MeshAttributeV8 {
+            usage: AttributeUsageV8::Normal,
+            data_type: AttributeDataTypeV8::Float2,
+            buffer_index: 1,
+            buffer_offset: 8,
+            sub_index: 3,
+        };
+
+        let attribute: MeshAttribute = (&attribute_v8).into();
+        assert_eq!("Normal3", attribute.name);
+        assert_eq!(DataType::Float2, attribute.data_type);
+        assert_eq!(1, attribute.index);
+        assert_eq!(8, attribute.offset);
+    }
+
     #[test]
     fn create_vertex_weights_mesh_v1_8() {
         // Version 1.8 uses an SsbhArray to store the weights.
@@ -1423,4 +1574,98 @@ mod tests {
         assert_eq!(12, get_size_in_bytes_v8(&AttributeDataTypeV8::Float3));
         assert_eq!(8, get_size_in_bytes_v8(&AttributeDataTypeV8::HalfFloat4));
     }
+
+    #[test]
+    fn max_influences_no_bones() {
+        assert_eq!(0, calculate_max_influences(&[]));
+    }
+
+    #[test]
+    fn max_influences_one_bone_no_weights() {
+        let influences = vec![BoneInfluence {
+            bone_name: "a".to_string(),
+            vertex_weights: Vec::new(),
+        }];
+        assert_eq!(0, calculate_max_influences(&influences));
+    }
+
+    #[test]
+    fn max_influences_one_bone() {
+        // Check that only influences are counted and not occurrences within an influence.
+        let influences = vec![BoneInfluence {
+            bone_name: "a".to_string(),
+            vertex_weights: vec![
+                VertexWeight {
+                    vertex_index: 0,
+                    vertex_weight: 0f32,
+                },
+                VertexWeight {
+                    vertex_index: 0,
+                    vertex_weight: 0f32,
+                },
+            ],
+        }];
+        // This is 1 and not 2 since there is only a single bone.
+        assert_eq!(1, calculate_max_influences(&influences));
+    }
+
+    #[test]
+    fn max_influences_three_bones() {
+        // Check that only influences are counted and not occurrences within an influence.
+        let influences = vec![
+            BoneInfluence {
+                bone_name: "a".to_string(),
+                vertex_weights: vec![
+                    VertexWeight {
+                        vertex_index: 0,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 0,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 0,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 3,
+                        vertex_weight: 0f32,
+                    },
+                ],
+            },
+            BoneInfluence {
+                bone_name: "b".to_string(),
+                vertex_weights: vec![
+                    VertexWeight {
+                        vertex_index: 2,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 1,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 3,
+                        vertex_weight: 0f32,
+                    },
+                ],
+            },
+            BoneInfluence {
+                bone_name: "c".to_string(),
+                vertex_weights: vec![
+                    VertexWeight {
+                        vertex_index: 0,
+                        vertex_weight: 0f32,
+                    },
+                    VertexWeight {
+                        vertex_index: 3,
+                        vertex_weight: 0f32,
+                    },
+                ],
+            },
+        ];
+
+        assert_eq!(3, calculate_max_influences(&influences));
+    }
 }
